/*    
       This software may be used and distributed according to the terms of
       the GNU General Public License (GPL), incorporated herein by
       reference.  Drivers based on this skeleton fall under the GPL and
       must retain the authorship (implicit copyright) notice.
 
       This program is distributed in the hope that it will be useful, but
       WITHOUT ANY WARRANTY; without even the implied warranty of
       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
       General Public License for more details.
 */
package cricketdaemon.position;


import cricketdaemon.beaconstat.*;
import cricketdaemon.beaconstat.BeaconDistStat.BeaconDist;
import cricketdaemon.module.*;
import cricketdaemon.clientlib.data.Position;
import cricketdaemon.reader.CartesianCoord;
import cricketdaemon.reader.parse.ParseMap;
import cricketdaemon.reader.sim.SimReader;
import cricketdaemon.space.*;
import cricketdaemon.space2.*;
import cricketdaemon.ServiceTime;
import cricketdaemon.UnitConv;
import cricketdaemon.Util;
import java.util.*;

/**
 * A module that implements a least-square matrix solver package to 
 * find the position of the device.
 *
 * As we will phase out the Space* modules, the PositionModule will
 * be modified to accept BeaconStatData instead of SpaceData.
 *
 * @author Allen Miu
 */
public class PositionModule extends ModuleThread 
    implements 
	BeaconStatData.BeaconStatDataHandler,
	SpaceData.SpaceDataHandler
{
    private static BitSet spaceFieldMask;



    /* PositionModule field types */
    public static final int CARTESIAN_COORD    = 1;
    public static final int DISTANCE_ESTIMATES = 2;
    public static final int POSERR             = 3;

    //public static final double K_NOMINAL = 
    //Util.SPEED_OF_SOUND/(Util.CLOCK_FREQ_HZ * 2.54);
    public static final double K_NOMINAL = UnitConv.cricketDistToCM(1);
    public static final int NUMBEACONS_REQUIRED = 3;

    private PositionSolver posSolver;

    private double stddevThresh = 15.0;
    private boolean aggressive = false;
    private boolean useEKF     = false;

    private EKFSolver ekfPVHSolver;
    private EKFSolver ekfPHSolver;
    
    private void initSolvers(double kthresh) {
	posSolver = new PositionSolver(K_NOMINAL, kthresh);
	ekfPVHSolver = new EKFSolver(false);
	ekfPHSolver  = new EKFSolver(true);
    }

    protected PositionModule(double kthresh) {
	initSolvers(kthresh);
    }

    public PositionModule(ModuleThread mt, double kthresh, double st)
    {
	this(mt, kthresh);
	stddevThresh = st;
    }

    public PositionModule(ModuleThread mt, double kthresh)
    {
	super("PositionModule", 0);
	prevBeacons = null;

	/*
	  Util.assert(mt instanceof Space2Module,
	  "PositionModule() only works with the Space2Module");
	*/

	BitSet desiredTypeMask = new BitSet();

	/*
	  if(mt instanceof Space2Module)
	  desiredTypeMask.set(SpaceModule.MODE);
	  else if(mt instanceof BeaconStatModule)
	  desiredTypeMask.set(BeaconStatModule.STAT);
	*/
	desiredTypeMask.set(BeaconStatModule.STAT);

	mt.register(this, desiredTypeMask);
	initSolvers(kthresh);
    }

    protected void moduleCleanUp() { return; }

    private ArrayList prevBeacons;

    protected boolean shouldUpdate(ArrayList a) { return true; }

    /* returns true iff all four distances have changed */
    /*
    protected boolean shouldUpdate(ArrayList a)
    {
	if(prevBeacons == null) {
	    prevBeacons = a;
	    return true;
	}

	int numSame = 0;
	for(int i = 0; i < a.size(); i++) {
	    for(int j = 0; j < prevBeacons.size(); j++) {
		if(a.get(i).equals(prevBeacons.get(j))) {
		    BeaconDistStat cur  = (BeaconDistStat) a.get(i);
		    BeaconDistStat prev = (BeaconDistStat) prevBeacons.get(j);
		    if(cur.distanceEst == prev.distanceEst)
			numSame++;
		}
	    }
	}
	if(numSame == 0) {
	    prevBeacons = a;
	    return true;
	}
	return false;
    }
    */

    /*
    private void pickBeacons(ArrayList beacons, int numBeacons)
    {
	int oldest = 0;
	long oldestTime = ((BeaconDistStat) beacons.get(0)).lastUpdate;
	for(int i = 1; i < numBeacons && i < beacons.size(); i++) {
	    BeaconDistStat b = (BeaconDistStat) beacons.get(i);
	    if(b.lastUpdate < oldestTime) {
		oldest = i;
		oldestTime = b.lastUpdate;
	    }
	}
	
        while(beacons.size() > numBeacons) {
	    BeaconDistStat b = 
		(BeaconDistStat) beacons.remove(numBeacons);
	    BeaconDistStat old = (BeaconDistStat) beacons.get(oldest);
	    if(b.lastUpdate > oldestTime) {
		beacons.remove(oldest);
		beacons.add(oldest, b);
		
		oldest = 0;
		oldestTime = ((BeaconDistStat) beacons.get(0)).lastUpdate;
		for(int i = 1; i < numBeacons && i < beacons.size(); i++) {
		    BeaconDistStat bb = (BeaconDistStat) beacons.get(i);
		    if(bb.lastUpdate < oldestTime) {
			oldest = i;
			oldestTime = bb.lastUpdate;
		    }
		}
	    }
	}
    }
    */

    private void printBeacons(ArrayList beacons)
    {
	for(int i = 0; i < beacons.size(); i++) {
	    System.out.println(beacons.get(i));
	}
    }

    private ArrayList rejects = new ArrayList();
    private void insertSortStddev(ArrayList l, BeaconDistStat bds) 
    {
	for(int i = 0; i < l.size(); i++) {
	    BeaconDistStat bl = (BeaconDistStat) l.get(i);
	    if(bds.stat.stddev > bl.stat.stddev) {
		l.add(i, bds);
		return;
	    }
	}
	l.add(bds);
    }
    private boolean filterBeacons(ArrayList beacons)
    {
	Iterator it = beacons.iterator();
	rejects.clear();
	while(it.hasNext()) {
	    BeaconDistStat b = (BeaconDistStat) it.next();
	    if(b.beacon.getField(ParseMap.CARTESIAN_COORD) == null) {
		//System.out.println("FILTERING:::: "+b);
		it.remove();
	    }
	    else if (K_NOMINAL*b.stat.stddev > 3*stddevThresh) {
		insertSortStddev(rejects, b);
	    }
	}
	int numRemove = (int) Math.min(beacons.size() - NUMBEACONS_REQUIRED, 
				       rejects.size());

	//System.out.println("numRemove "+numRemove+ " "+beacons.size()+" "+rejects.size());

	while(numRemove-- > 0)
	    beacons.remove(beacons.indexOf(rejects.remove(rejects.size()-1)));


	return beacons.size() >= NUMBEACONS_REQUIRED;
    }


    private void _send(PositionData newEstimate) 
    {
	if(newEstimate != null) 
	    send(newEstimate);
	else 
	    sendErr();
    }

    public void processBeaconStatData(BeaconStatData newData, BitSet fieldMask)
    {
	PositionData newEstimate = null;

	/*
	  if(useEKF && !aggressive)
	  newEstimate = ekf(newData, fieldMask);
	  else
	  newEstimate = llsq(newData, fieldMask);
	*/
	  /* send position data */
	//_send(newEstimate);	
	
	_send(ekf(newData, fieldMask));
	_send(llsq(newData, fieldMask));
    }

    private PositionData ekf(BeaconStatData newData, BitSet fieldMask)
    {
	//System.out.println("foobar "+dist+" "+(tup.distances.size()));
	double      fCovPVH = 0;
	double      fCovPH  = 0;

	PositionData posPVH = runEKF(newData, fieldMask, ekfPVHSolver);
	if(posPVH != null)
	    fCovPVH = ekfPVHSolver.getfCov();

	PositionData posPH = runEKF(newData, fieldMask, ekfPHSolver);
	if(posPVH != null)
	    fCovPH = ekfPHSolver.getfCov();

	double wPVH = fCovPVH/(fCovPVH + fCovPH);
	double wPH  = fCovPH/(fCovPVH + fCovPH);


	EKFSolver baseEKF = null;
	CartesianCoord coordPVH = null;
	CartesianCoord coordPH  = null;
	
	// PositionData requires "meta info" such as beacon coordinates, etc
	// embedded in the ekfSolver.  We can use either ekfSolver to generate
	// the necessary metadata as long as their fCov is not zero (bad state)
	if(posPH != null) {
	    baseEKF = ekfPHSolver;
	    coordPH = posPH.getCoord();
	    //System.out.println("fCovPH="+fCovPH+" pos="+coordPH);
	}
	// PVH takes precedent
	if(posPVH != null) {
	    baseEKF = ekfPVHSolver;
	    coordPVH = posPVH.getCoord();
	    //System.out.println("fCovPVH="+fCovPVH+" pos="+coordPVH);
	}

	Position wPos = Position.weightSum(wPVH, coordPVH, wPH, coordPH);
	PositionData result = null;

	// baseEKF == null ---> wPos.x = 0 wPos.y = 0 wPos.z = 0
	if(baseEKF != null) {
	    result = baseEKF.curPos(wPos.x, wPos.y, wPos.z);
	}

	if(result != null)
	    result.setAlgor("ekf-passive");
	return result;
    }

    private PositionData runEKF(BeaconStatData newData, BitSet fieldMask, EKFSolver ekfSolver)
    {
	BeaconDistStat tup = newData.latestStat;
	CartesianCoord c = (CartesianCoord) 
	    tup.beacon.getField(ParseMap.CARTESIAN_COORD);

	if(c == null)
	    return null;

	Integer dist = 
	    ((BeaconDist) tup.distances.get(tup.distances.size()-1)).dist;

	// pre-condition:  newData.beacon is the raw reading of the last tup.distances
	PositionData result =
	    ekfSolver.process(newData.beacon, tup.space, tup.id, 
				 c, dist.intValue()*K_NOMINAL);
	if(result == null) {
	    //System.out.println("llsq resetting!");
	    result = llsq(newData, fieldMask);
	    if(result != null)
		ekfSolver.reset(result);
	}	

	return result;
    }

    private PositionData llsq(BeaconStatData newData, BitSet fieldMask)
    {

	// values array needs to cloned as it is shared by multiple modules
	ArrayList beacons = (ArrayList) newData.values.clone();
	
	if(beacons == null) {
	    Util.warn("PositionModule::processBeaconStatData() beacons == null");
	    return null;
	}

	if(!filterBeacons(beacons)) {
	    Util.log("PositionModule: Too few valid beacons \n\t("+
		     "stddev too high or coord not available)");
	    return null;
	}

	if(!shouldUpdate(beacons))
	    return null;

	//int size = (beacons.size() < 4) ? beacons.size() : 4;
	int size = beacons.size();
	
	double[] x     = new double[size];
	double[] y     = new double[size];
	double[] z     = new double[size];
	String[] space = new String[size];
	int[]    id    = new int[size];
	double[] dist  = new double[size];

	Iterator it = beacons.iterator();

	/*
	  System.out.println("********* before ********* ");
	  printBeacons(beacons);
	  pickBeacons(beacons);
	  System.out.println("****** after ****** ");
	  printBeacons(beacons);
	*/

	//pickBeacons(beacons, size);

	int i = 0;
	while(it.hasNext()) {
	    BeaconDistStat b = (BeaconDistStat) it.next();
	    CartesianCoord c = (CartesianCoord) 
		b.beacon.getField(ParseMap.CARTESIAN_COORD);

	    if(c != null) {
		x[i]     = c.x;
		y[i]     = c.y;
		z[i]     = c.z;
		space[i] = b.space;
		id[i]    = b.id;

		//dist[i]  = b.distanceEst;


		if(aggressive)
		    dist[i] = ((BeaconDist) b.distances.get(b.distances.size()-1)).dist.doubleValue();
		else
		    dist[i]  = b.distanceEst;


		//Util.log(x[i]+" "+y[i]+" "+z[i]+" "+space[i]+
		//" "+id[i]+" "+dist[i]);
		 System.out.println(x[i]+" "+y[i]+" "+z[i]+" "+
		 space[i]+" "+id[i]+" "+dist[i]);
		i++;
	    }
	}
	
	/*
	  for(int combo = 0; combo < comboIndexA.length; combo++) {
	      	    System.out.println("combo: "+combo);
	  if(combo == 8)
	  newEstimate = positionEstimate(x, y, z, space, id, dist, combo);
	  else
	  positionEstimate(x, y, z, space, id, dist, combo);
	  }
	*/

	PositionData newEstimate = 
	    posSolver.positionEstimateVKnown(newData.beacon, x, y, z, space, id, dist);
	if(newEstimate != null)
	    newEstimate.setAlgor("llsq-passive");
	return newEstimate;

	/*
	if(size <= 4) 
	    newEstimate = positionEstimateVKnown(x, y, z, space, id, dist);
	else {
	    newEstimate = positionEstimateVUnknown(x, y, z, space, id, dist);
	}
	*/


    }

    public void processSpaceData(SpaceData newData, BitSet fieldMask)
    {
	return;
    }

    /*
    // fixme:  may want to discard old data that has been
    // queued up and return the position from the last queued item
    public void processSpaceData(SpaceData newData, BitSet fieldMask)
    {
	//int seqNum = newData.seqNum;
	DistanceEstimate[] estimates = newData.getMode();
	SpaceBeaconTuple[] tuples = convertToTuples(estimates);

	if(tuples == null || tuples.length < NUMBEACONS_REQUIRED) {
	    Util.log("PositionModule::processSpaceData() not enough beacons "+
		     "that have coordinate infomration");
	    return;
	}

	double[] x     = new double[NUMBEACONS_REQUIRED];
	double[] y     = new double[NUMBEACONS_REQUIRED];
	double[] z     = new double[NUMBEACONS_REQUIRED];
	String[] space = new String[NUMBEACONS_REQUIRED];
	int[]    id    = new int[NUMBEACONS_REQUIRED];
	double[] dist  = new double[NUMBEACONS_REQUIRED];


	// FIXME: need strategy to pick a subset of required beacons 
	for(int i = 0; i < NUMBEACONS_REQUIRED; i++) {
	    Util.assert_nonjava(tuples[i].coord != null,
			"PositionModule::processSpaceData() need to run with"+
			" beacon coordinate mapping on");
	    x[i]     = tuples[i].coord.x;
	    y[i]     = tuples[i].coord.y;
	    z[i]     = tuples[i].coord.z;
	    space[i] = tuples[i].space;
	    id[i]    = tuples[i].beaconID;
	    dist[i]  = tuples[i].distance;
	    
	    Util.log("Beacon ["+i+"]: "+x[i]+" "+y[i]+" "+dist[i]);
	}

	PositionData newEstimate = positionEstimate(x, y, z, 
						    space, id, dist, 
						    8);
	
	// send position data 
	send(newEstimate);
    }
    */

    private void send(PositionData newEstimate) 
    {
	BitSet positionFieldMask = (BitSet) CLEARMASK.clone();
	positionFieldMask.set(CARTESIAN_COORD);
	positionFieldMask.set(DISTANCE_ESTIMATES);
	//System.out.println("PositionModule::process() "+newEstimate);
	transfer(newEstimate, positionFieldMask);
	
    }

    private void sendErr()
    {
	BitSet positionFieldMask = (BitSet) CLEARMASK.clone();
	positionFieldMask.set(POSERR);
	PositionData.invalidInstance.beacon.timeStamp = new Long(System.currentTimeMillis());
	PositionData.invalidInstance.beacon.timeGenerated = 
	    PositionData.invalidInstance.beacon.timeStamp;
	transfer(PositionData.invalidInstance, positionFieldMask);
    }

    public SpaceBeaconTuple[] convertToTuples(DistanceEstimate[] d) 
    {
	SpaceBeaconTuple[] result = new SpaceBeaconTuple[d.length];
	for(int i = 0; i < result.length; i++)
	    result[i] = (SpaceBeaconTuple) d[i];
	return result;
    }

    // cmd value0 value1 ...
    // aggressive [true|false]
    // stddevthresh double
    // useEKF [true|false]
    // for now, no need synchronized lock
    public void parseCmd(StringTokenizer indata) 
    {
	int numTokens = indata.countTokens();

	if(numTokens < 1)
	    return;

	String cmd = indata.nextToken();
	if(cmd.equals("aggressive")) {
	    String s = null;
	    if(numTokens == 2) {
		s = indata.nextToken();
		aggressive = Boolean.valueOf(s).booleanValue();
	    }
	    //System.out.println("foobar aggressive "+aggressive+" str="+s+"E");
	    return;
	}

	if(cmd.equals("useEKF")) {
	    String s = null;
	    if(numTokens == 2) {
		s = indata.nextToken();
		useEKF = Boolean.valueOf(s).booleanValue();
	    }
	    //System.out.println("foobar useEKF "+useEKF+" str="+s+"E");
	    return;
	}

	if(cmd.equals("stddevthresh")) {
	    if(numTokens == 2) {
		stddevThresh = Double.parseDouble(indata.nextToken());
	    }
	    //System.out.println("foobar stddevthresh "+stddevThresh);

	    return;
	}

    }


}




